/******************************************************
STM8S003K3 16MHz, IAR1.4+STM8S_StdPeriph_Lib_V2.1.0
******************************************************/
#include "stm8s.h"
#include "iostm003.c"
#include <stdbool.h>

#define A       PD_ODR_ODR5
#define B       PD_ODR_ODR4
#define C       PD_ODR_ODR3
#define D       PD_ODR_ODR2
#define E       PD_ODR_ODR1
#define F       PB_ODR_ODR6
#define G       PB_ODR_ODR7
#define DP      PF_ODR_ODR4
#define N0      PA_ODR_ODR3
#define N1      PD_ODR_ODR7
#define N2      PD_ODR_ODR6 

#define therm0	PB_IDR_IDR2
#define therm1	PB_IDR_IDR3
#define key0    PB_IDR_IDR4
#define key1	PB_IDR_IDR5

#define REL     PE_ODR_ODR5

#define ratio 100
#define lres  20   //Ohms
#define ttr   (33*10)/6 //55
#define I_max_adc (50UL*ratio*ttr)/lres //=I_max*100 137(,)50

#define R_upr 2000 //Ohms
#define R_lwr 1000   //Ohms
#define U_max_adc ((500UL*(R_upr+R_lwr))/R_lwr) //=U_max*100, 15V*100

#define MaxAmp 120
#define MinAmp 20

enum btn_states {prs, rls, hold} btn0,btn1,btn2;
enum dmodes {regI,run,loaux,ocp,astick,ovl,sdn,trm} dev_st=run;

s16 def_tmp;


u8 tkey[2],tmr[4];
u16 tim[2];


u8 dbuf[3]={0,0,2};
u8 dpp=0;

u16 Idef;
s16 I_tmp;
union {u8 adc[2]; u16 adcw;} adcu1,adcu0;
u16 adc1val, adc0val;

u8 x;

#define msize 20
__no_init u8 eeIval [msize] @ 0x4000;

inline static void gpio_cfg()
{
  PB_DDR=PC_DDR=PD_DDR=PE_DDR=PF_DDR=0xff; //conf as output
  PA_CR1=PB_CR1=PC_CR1=PD_CR1=PE_CR1=PF_CR1=0xff; //push-pull|pullup

  PA_ODR = 0x0f-8-1; PA_DDR = 0x0f-1;
  PB_ODR = 1<<3|1<<2; PB_DDR = 1<<7|1<<6; PB_CR1&= ~(1<<1|1);
  PC_ODR = 1<<7|1<<4|1<<3|1<<2|1<<1;
  PD_DDR = 0xff-1; //BKIN

}
inline static void clk_cfg()
{
  CLK_DeInit();
  CLK->CKDIVR = 0;     //16MHz
  CLK->PCKENR1 = CLK_PCKENR1_TIM1|CLK_PCKENR1_TIM2;
}
inline static void t1_cfg()
{
  {
    #define toppwm 0xff //31250 Hz

    TIM1->ARRH = 0;
    TIM1->ARRL = toppwm;

    TIM1->CCR1H=TIM1->CCR2H=0;
    TIM1->CCR1L=TIM1->CCR2L=0;
    TIM1->CCR3H=TIM1->CCR4H=0;
    TIM1->CCR3L=TIM1->CCR4L=toppwm;

  }

    TIM1->CR1 = TIM1_COUNTERMODE_CENTERALIGNED1|TIM1_CR1_ARPE|TIM1_CR1_URS;
    TIM1->CR2 = TIM1_CR2_CCPC;
    TIM1->RCR = 1;

    TIM1->CCMR1 = TIM1->CCMR2 = TIM1_OCMODE_PWM1|TIM1_CCMR_OCxPE;
    TIM1->CCMR3 = TIM1->CCMR4 = TIM1_OCMODE_PWM2|TIM1_CCMR_OCxPE;
    TIM1->CCER1 = TIM1_CCER1_CC1E|TIM1_CCER1_CC2E;
    TIM1->CCER2 = TIM1_CCER2_CC3E|TIM1_CCER2_CC4E;
    TIM1->EGR  |= TIM1_EGR_COMG;
    TIM1->IER   = TIM1_IER_UIE;

    TIM1->BKR  = TIM1_BKR_BKE;
    TIM1->OISR = 0x55;
    TIM1->CR1 |= TIM1_CR1_CEN;
    TIM1->BKR |= TIM1_LOCKLEVEL_3;
}
static inline void t2_cfg()
{
  TIM2->PSCR  = 1;
  TIM2->ARRH  = 6;
  TIM2->ARRL  = 0x40; //5000Hz
  TIM2->CR1   = TIM2_CR1_ARPE|TIM2_CR1_URS;
  TIM2->IER   = TIM2_IER_UIE;

  TIM2->CR1 |= TIM2_CR1_CEN;
}
inline static void adc_cfg()
{
  ADC1->CR1 = ADC1_PRESSEL_FCPU_D18;
  ADC1->CR2 = ADC1_ALIGN_RIGHT|ADC1_CR2_SCAN;
  ADC1->CR3 = ADC1_CR3_DBUF;
  ADC1->CSR = 1; //channel sel
  ADC1->TDRL = 3; //smtt trg on adc pins is off
  ADC1->CR1 |= ADC1_CR1_ADON;
}
void iwdg_cfg()
{
IWDG_WriteAccessCmd(IWDG_WriteAccess_Enable);
IWDG_SetPrescaler(IWDG_Prescaler_4);
IWDG_SetReload(0);
IWDG_Enable();
IWDG_ReloadCounter();
}

void scan_adc()
{

  if(ADC1->CSR&ADC1_CSR_EOC)
  {
	adcu0.adc[1] = ADC1->DB0RL;
	adcu0.adc[0] = ADC1->DB0RH;
	adcu1.adc[1] = ADC1->DB1RL;
	adcu1.adc[0] = ADC1->DB1RH;

	ADC1->CSR = 1; ADC1->CR1 |= ADC1_CR1_ADON;

	#define ADC_itn 16

	u8 i;
	static u16 ADC0_tmp[ADC_itn],ADC1_tmp[ADC_itn];

	for (i=0; i<(ADC_itn-1); i++ )
	{
	  ADC0_tmp[i] = ADC0_tmp[(u16)i+1];
	  ADC1_tmp[i] = ADC1_tmp[(u16)i+1];
	}
	ADC0_tmp[ADC_itn-1] = adcu0.adcw;
	ADC1_tmp[ADC_itn-1] = adcu1.adcw;
	adc1val = adc0val = 0;
	for (i=0; i<(ADC_itn); i++)
	{
	  adc0val += ADC0_tmp[i];
	  adc1val += ADC1_tmp[i];
	}
	adc0val /= ADC_itn;
	adc1val /= ADC_itn;
  }
}

void convert_data (u16 d)
{
u8 k;
  for (k=0;k<3;k++)
  {
   dbuf[k]=d%10;
   d/=10;
  }
}
void display_data ()
{
  static u8 pc;

  (pc)?(pc--):(pc = 2);
  N0=N1=N2=0; DP=1;

  switch (dbuf[pc])
 {
  case 0: A=0; B=0; C=0; D=0; E=0; F=0; G=1; break;
  case 1: A=1; B=0; C=0; D=1; E=1; F=1; G=1; break;
  case 2: A=0; B=0; C=1; D=0; E=0; F=1; G=0; break;
  case 3: A=0; B=0; C=0; D=0; E=1; F=1; G=0; break;
  case 4: A=1; B=0; C=0; D=1; E=1; F=0; G=0; break;
  case 5: A=0; B=1; C=0; D=0; E=1; F=0; G=0; break;
  case 6: A=0; B=1; C=0; D=0; E=0; F=0; G=0; break;
  case 7: A=0; B=0; C=0; D=1; E=1; F=1; G=1; break;
  case 8: A=0; B=0; C=0; D=0; E=0; F=0; G=0; break;
  case 9: A=0; B=0; C=0; D=0; E=1; F=0; G=0; break;
  case ' ': A=1; B=1; C=1; D=1; E=1; F=1; G=1; break;
  case 'u': A=1; B=1; C=0; D=0; E=0; F=1; G=1; break;
  case 'c': A=1; B=1; C=1; D=0; E=0; F=1; G=0; break;
  case 'b': A=1; B=1; C=0; D=0; E=0; F=0; G=0; break;
  case 't': A=1; B=1; C=1; D=0; E=0; F=0; G=0; break;
  case 'l': A=1; B=1; C=1; D=0; E=0; F=1; G=1; break;
  case 'o': A=1; B=1; C=0; D=0; E=0; F=1; G=0; break;
  case 'A': A=0; B=0; C=0; D=1; E=0; F=0; G=0; break;
  case 'H': A=1; B=0; C=0; D=1; E=0; F=0; G=0; break;
  case 'C': A=0; B=1; C=1; D=0; E=0; F=0; G=1; break;
  case 'P': A=0; B=0; C=1; D=1; E=0; F=0; G=0; break;
  case 'E': A=0; B=1; C=1; D=0; E=0; F=0; G=0; break;
  case 'U': A=1; B=0; C=0; D=0; E=0; F=0; G=1; break;
  case 'F': A=0; B=1; C=1; D=1; E=0; F=0; G=0; break;
  case 'L': A=1; B=1; C=1; D=0; E=0; F=0; G=1; break;
  case 'r': A=1; B=1; C=1; D=1; E=0; F=1; G=0; break;
  case '_': A=1; B=1; C=1; D=0; E=1; F=1; G=1; break;
  case 'd': A=1; B=0; C=0; D=0; E=0; F=1; G=0; break;
  case 'n': A=1; B=1; C=0; D=1; E=0; F=1; G=0; break;
  case '': A=0; B=0; C=1; D=1; E=1; F=0; G=1; break;
  case 'O': A=0; B=0; C=0; D=0; E=0; F=0; G=1; break;
  default:  A=1; B=1; C=1; D=1; E=1; F=1; G=0; break; //-
 }

   switch (pc)
 {
  case 0: N0=1; if(dpp == 0) {DP=0;} break;
  case 1: N1=1; if(dpp == 1) {DP=0;} break;
  case 2: if(dbuf[2]){N2=1;} if(dpp == 2) {DP=0;} break;
  default: DP=1; break;
 }
 }

void scan_keys ()
{
#define lp  4
#define lh  10
#define lr  2

static u8 count[6];

if ((!key0)&&(!tkey[0])) {count[0]++;tkey[0]=lr;}
  else if((key0)&&(!tkey[0])) {count[0]=count[1]=0;btn0=rls;}
if ((count[0]>lp)&&(btn0!=hold)) {btn0=prs; count[0]=0; count[1]++;}
//if (count[1]>lh) {btn0=hold; count[1]=0;}

if ((!key1)&&(!tkey[1])) {count[2]++;tkey[1]=lr;}
  else if((key1)&&(!tkey[1])) {count[3]=count[2]=0;btn1=rls;}
if ((count[2]>lp)&&(btn1!=hold)) {btn1=prs; count[2]=0; count[3]++;}
//if (count[3]>lh) {btn1=hold; count[3]=0;}
/*
if ((!key2)&&(!tkey[2])) {count[4]++;tkey[2]=lr;}
  else if((key2)&&(!tkey[2])) {count[5]=count[4]=0;btn2=rls;}
if ((count[4]>0xf0)&&(btn2!=hold)) {btn2=prs; count[4]=0; count[5]++;}
if (count[5]>lh) {btn2=hold; count[5]=0;}
*/
}

struct encoder_struct
{
 s16* value;
 s16  val_min;
 s16  val_max;
 u8   step;
 u8   cycle_form;
}encoder_data={&def_tmp,0,100,1,0};
void scan_enc (struct encoder_struct sst)
{
u8 New; bool a1,a0;
static u8 EncState;

if(btn0==rls) a0=1; else a0=0;
if(btn1==rls) a1=1; else a1=0;
New = (a1<<1)+a0;



switch(EncState)
{
case 0:
{
if(New == 2)
  if(*sst.value<sst.val_max-sst.step)*sst.value+=sst.step;
  else if(!sst.cycle_form)*sst.value=sst.val_min;
  else *sst.value=sst.val_max;
if(New == 1)
  if(*sst.value>sst.val_min+sst.step)*sst.value-=sst.step;
  else if(!sst.cycle_form)*sst.value=sst.val_max;
  else *sst.value=sst.val_min;
break;
}
case 1:
{
if(New == 0)
  if(*sst.value<sst.val_max-sst.step)*sst.value+=sst.step;
  else if(!sst.cycle_form)*sst.value=sst.val_min;
  else *sst.value=sst.val_max;
if(New == 3)
  if(*sst.value>sst.val_min+sst.step)*sst.value-=sst.step;
  else if(!sst.cycle_form)*sst.value=sst.val_max;
  else *sst.value=sst.val_min;
break;
}
case 2:
{
if(New == 3)
  if(*sst.value<sst.val_max-sst.step)*sst.value+=sst.step;
  else if(!sst.cycle_form)*sst.value=sst.val_min;
  else *sst.value=sst.val_max;
if(New == 0)
  if(*sst.value>sst.val_min+sst.step)*sst.value-=sst.step;
  else if(!sst.cycle_form)*sst.value=sst.val_max;
  else *sst.value=sst.val_min;
break;
}
case 3:
{
if(New == 1)
  if(*sst.value<sst.val_max-sst.step)*sst.value+=sst.step;
  else if(!sst.cycle_form)*sst.value=sst.val_min;
  else *sst.value=sst.val_max;
if(New == 2)
  if(*sst.value>sst.val_min+sst.step)*sst.value-=sst.step;
  else if(!sst.cycle_form)*sst.value=sst.val_max;
  else *sst.value=sst.val_min;
break;
}
}

EncState = New;
}

u8 read_ee_i()
 {
  for(u8 i=0;i<msize;i++)
   {
     if (eeIval[i]!=0xff) return eeIval[i];
   }
  return 0xff;
 }
void write_ee_i (u8 databyte)
{
 for(u8 i=0;i<msize;i++)
 {
   if(eeIval[i]==databyte) return;

   if(eeIval[i]!=0xff)
    {
      FLASH_Unlock(FLASH_MEMTYPE_DATA);
      if(i<msize-1)
      {
       eeIval[i+1]=databyte;
       eeIval[i]=0xff;
      }
     else
      {
       eeIval[0]=databyte;
       eeIval[i]=0xff;
      }
      FLASH_Lock(FLASH_MEMTYPE_DATA);
      return;
    }
   else if(i==msize-1)
    {
      FLASH_Unlock(FLASH_MEMTYPE_DATA);
      eeIval[0]=databyte;
      FLASH_Lock(FLASH_MEMTYPE_DATA);
    }
 }

}

void devcon(),pwm_on(),pwm_off(),set_pwm(u8 someval),stab_process(u8 Idef);

INTERRUPT_HANDLER(TIM1_X3, 11)
{

  TIM1->CCR1L=TIM1->CCR2L=x;
  TIM1->CCR3L=TIM1->CCR4L = toppwm - TIM1->CCR1L;

  IWDG_ReloadCounter();
  TIM1->SR1 = 0;
}

INTERRUPT_HANDLER(TIM2_OVF, 13)
{
 scan_keys ();
 scan_enc (encoder_data);

  if(tkey[0])tkey[0]--;
  if(tkey[1])tkey[1]--;
  if(tmr[0])tmr[0]--;
  if(tmr[1])tmr[1]--;
  if(tmr[2])tmr[2]--;
  if(tmr[3])tmr[3]--;
  if(tim[0])tim[0]--;
  if(tim[1])tim[1]--;

  TIM2->SR1 = 0;
}

void main()
{
gpio_cfg();
clk_cfg();
adc_cfg();
ITC_SetSoftwarePriority(ITC_IRQ_TIM1_OVF, ITC_PRIORITYLEVEL_3);
ITC_SetSoftwarePriority(ITC_IRQ_TIM2_OVF, ITC_PRIORITYLEVEL_1);
t1_cfg();
t2_cfg();


if((!read_ee_i())||(read_ee_i()>160))write_ee_i(60);

ADC1->CR1 |= ADC1_CR1_ADON;

iwdg_cfg();
asm("rim");

  while(1)
  {
	devcon();
	display_data();
	scan_adc();

  }
}

void assert_failed(uint8_t* file, uint32_t line){}

void devcon()
 {
  static bool itr=0;
  static bool ovl_trg=0;



  switch(dev_st) 
  {
   case run:
   {
    static u8 t;
    if(!itr)
    {
    itr=1; tmr[0]=0xff; t=80; //4s
    dbuf[2]='r';dbuf[1]='u';dbuf[0]='n';dpp=5;
    pwm_off(); REL=0;
    }

    if((!tmr[0])&&(t)){tmr[0]=0xff;t--;}
    else if(!t) {if(adc0val>(((u32)12*0x3ff*100)/U_max_adc)) dev_st=regI;}

   if(dev_st!=run){itr=0;}
   } break;

   case regI:
   {
    static u8 acn, moe, sdn_cnt, ee_cnt, therm_cnt;
    static bool i_def_trg, sdn_trg;

    if(!itr)
    {
    itr=1; acn=10; moe=ee_cnt=0; i_def_trg=sdn_cnt=sdn_trg=0;
    Idef=I_tmp=read_ee_i(); REL=1;
    dpp=5; therm_cnt=10;
    encoder_data.val_max=MaxAmp;
    encoder_data.val_min=MinAmp;
    encoder_data.step=1;
    encoder_data.cycle_form=1;
    encoder_data.value = &I_tmp;
    }

    stab_process(I_tmp);

    if(!tim[0])
     {
       tim[0]=500;
       convert_data(I_tmp); //  1/10c   .

       if((I_tmp!=Idef)&&(!i_def_trg)) {ee_cnt=50;i_def_trg=1;}
       else if((I_tmp!=Idef)&&(i_def_trg)&&(!ee_cnt)&&(!sdn_trg))
              {Idef=I_tmp;i_def_trg=0;if(I_tmp>MinAmp+2)write_ee_i(Idef);}
       else if((I_tmp==Idef)&&i_def_trg&&(!ee_cnt)) i_def_trg=0;  //    5    
       else if(ee_cnt)ee_cnt--;
       
       if((I_tmp<MinAmp+2)&&(!sdn_trg)){sdn_trg=1; sdn_cnt=40;} //ec       4 
       else if((I_tmp<MinAmp+2)&&(!sdn_cnt)&&sdn_trg){sdn_trg=0; dev_st=sdn;}
       else if((I_tmp<MinAmp+2)&&(sdn_cnt)){sdn_cnt--;}
       else if(I_tmp!=MinAmp) sdn_trg=0;
       
       if((!therm0||!therm1)&&therm_cnt) therm_cnt--;
       else if ((!therm0||!therm1)&&!therm_cnt) dev_st=trm;
       else if (therm0&&therm1)therm_cnt = 10;
     }



    if((adc0val<((12*0x3ff)/(U_max_adc/100)))&&(acn)&&(!tmr[1])) {acn --;tmr[1]=0x0f;}
    else if ((adc0val>((13*0x3ff)/(U_max_adc/100)))&&(!tmr[1])) acn=10; //  

    if(TIM1->BKR&TIM1_BKR_MOE){moe=0;}
    else if(moe<0x0f){moe++;pwm_on();}
    else {dev_st=ocp;} //     

    if (!acn) dev_st=loaux;

    if(dev_st!=regI){itr=0;}
   } break;

   case ovl:
   {

    if(!itr)
    {
     itr=1; ovl_trg=0;
     dbuf[2]='o';dbuf[1]='u';dbuf[0]='l';dpp=3;
     pwm_off(); REL=0;
     tim[1] = 0xffff;
     I_tmp=Idef=MinAmp;
    }

    if(!tim[1])
    {
      tim[1]=5000;

      if((I_tmp>MaxAmp/5)&&(!ovl_trg)) {ovl_trg=1;dpp=0;}
      else if((I_tmp>MaxAmp/5)&&(ovl_trg)){ovl_trg=0;dev_st=run;}
      else if((I_tmp<MaxAmp/5)&&(ovl_trg)){ovl_trg=0;I_tmp=MinAmp;dpp=3;}

    }

    if(dev_st!=ovl){itr=0;}
   } break;

   case loaux:
   {
    if(!itr)
    {
     itr=1;
     dbuf[2]='l';dbuf[1]='o';dbuf[0]='H';dpp=3;
     pwm_off(); REL=0;
     tim[1] = 0xffff;
     I_tmp=Idef=MinAmp;
    }

    if(!tim[1])
    {
      tim[1]=5000;

      if((I_tmp>MaxAmp/5)&&(!ovl_trg)) {ovl_trg=1;dpp=0;}
      else if((I_tmp>MaxAmp/5)&&(ovl_trg)){ovl_trg=0;dev_st=run;}
      else if((I_tmp<MaxAmp/5)&&(ovl_trg)){ovl_trg=0;I_tmp=MinAmp;dpp=3;}

    }

    if(dev_st!=loaux){itr=0;}
   } break;

   case ocp:
   {
    if(!itr)
    {
     itr=1;
     dbuf[2]='O';dbuf[1]='C';dbuf[0]='P';dpp=3;
     pwm_off(); REL=0;
     tim[1] = 0xffff;
     I_tmp=Idef=MinAmp;
    }

    if(!tim[1])
    {
      tim[1]=5000;

      if((I_tmp>MaxAmp/5)&&(!ovl_trg)) {ovl_trg=1;dpp=0;}
      else if((I_tmp>MaxAmp/5)&&(ovl_trg)){ovl_trg=0;dev_st=run;}
      else if((I_tmp<MaxAmp/5)&&(ovl_trg)){ovl_trg=0;I_tmp=MinAmp;dpp=3;}

    }

    if(dev_st!=ocp){itr=0;}
   } break;

   case sdn:
   {
    if(!itr)
    {
     itr=1;
     dbuf[2]='_';dbuf[1]='_';dbuf[0]='_';dpp=5;
     pwm_off(); REL=0;
     I_tmp=Idef=MinAmp;
    }

    if(!tim[1])
    {
      tim[1]=5000;

      if((I_tmp>MaxAmp/5)&&(!ovl_trg)) {ovl_trg=1;dpp=0;}
      else if((I_tmp>MaxAmp/5)&&(ovl_trg)){ovl_trg=0;dev_st=run;}
      else if((I_tmp<MaxAmp/5)&&(ovl_trg)){ovl_trg=0;I_tmp=MinAmp;dpp=3;}

    }

    if(dev_st!=sdn){itr=0;}
   } break;

   case trm:
    {
    if(!itr)
    {
     itr=1;
     dbuf[2]='H';dbuf[1]='t'; dpp=1;
     if(!therm0)dbuf[0]='O';
     else if(!therm1) dbuf[0]='1';
     pwm_off(); REL=0;
    }    
    
    if(therm0&&therm1&&(!tim[1])) dev_st=run; else tim[1]=0xffff;
     if(dev_st!=trm){itr=0;}      
    } break;
   
   default:break;
   }
 }

void pwm_on()
 {
  TIM1->BKR |= TIM1_BKR_MOE;
 }
void pwm_off()
 {
  TIM1->BKR &= ~TIM1_BKR_MOE;
 }

void set_pwm(u8 someval)
{
#define dtv 20 //*62ns
u8 tmp;

tmp=(u8)(((u16)someval*(toppwm/2-dtv))>>8);
if(tmp>dtv)x=tmp;
else x=0;

}

void stab_process(u8 Idef)
{
 static u8 pwm_factor;
 const u16 f=I_max_adc;
 u8 i_fact;
 static u8 i_tmp=0;
 static bool itrg=0;


 i_fact=((u32)adc1val*f)/(0x3ffUL*100); //, max 25*5,5 137A

 if((i_fact<MinAmp)&&(i_tmp>=i_fact)&&(!tim[1])) 
  {pwm_factor=(u16)dtv*2*0xff/(toppwm/2-dtv); itrg=1; dpp=0;}//      ,  
 //else if((i_fact<MinAmp)&&(tim[1])){pwm_factor=(u8)(((u32)0xff*Idef*100)/f);} 
 else if(itrg){itrg=0; pwm_factor=(u8)(((u32)0xff*Idef*100)/f);dpp=5;}//    
 else if(i_fact<Idef) //   
 {
 if(pwm_factor<0xff){pwm_factor++;dpp=5;tim[1]=0xffff;}
 }
 else if((i_fact>Idef)&&(pwm_factor>0xff*Idef/f)) //   
 {
  pwm_factor--;dpp=5;tim[1]=0xffff;
 }


 set_pwm(pwm_factor);

 if(!tmr[3]){tmr[3]=0xff/2; i_tmp=i_fact;}
}
